#include "imagepng.h"

imagepng::imagepng() {

    row = NULL;
    width  = 0;
    height = 0;
}

// Удаление из памяти
imagepng::~imagepng() {
    destroy();
}

// Загрузка PNG в память
imagepng::imagepng(const char* filename) {
    load(filename);
}

// Удалить из память
void imagepng::destroy() {

    if (row) {

        for (int y = 0; y < height; y++) {
            free(row[y]);
        }

        free(row);
    }
}

// Загрузка файла PNG
// https://gist.github.com/niw/5963798
void imagepng::load(const char* filename) {

    FILE* fp = fopen(filename, "rb");

    if (fp == NULL) { printf("File not found: %s\n", filename); exit(1); }

    destroy();

    png_structp png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png) abort();

    png_infop info = png_create_info_struct(png);
    if (!info) abort();

    if (setjmp(png_jmpbuf(png))) abort();

    png_init_io  (png, fp);
    png_read_info(png, info);

    width      = png_get_image_width(png, info);
    height     = png_get_image_height(png, info);
    color_type = png_get_color_type(png, info);
    bit_depth  = png_get_bit_depth(png, info);

    // Read any color_type into 8bit depth, RGBA format.
    // See http://www.libpng.org/pub/png/libpng-manual.txt

    if (bit_depth == 16) {
        png_set_strip_16(png);
    }

    if (color_type == PNG_COLOR_TYPE_PALETTE) {
        png_set_palette_to_rgb(png);
    }

    // PNG_COLOR_TYPE_GRAY_ALPHA is always 8 or 16bit depth.
    if(color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8) {
        png_set_expand_gray_1_2_4_to_8(png);
    }

    if (png_get_valid(png, info, PNG_INFO_tRNS)) {
        png_set_tRNS_to_alpha(png);
    }

    // These color_type don't have an alpha channel then fill it with 0xff.
    if (color_type == PNG_COLOR_TYPE_RGB ||
        color_type == PNG_COLOR_TYPE_GRAY ||
        color_type == PNG_COLOR_TYPE_PALETTE) {
            png_set_filler(png, 0xFF, PNG_FILLER_AFTER);
    }

    if (color_type == PNG_COLOR_TYPE_GRAY || color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
        png_set_gray_to_rgb(png);
    }

    png_read_update_info(png, info);

    if (row) abort();

    row = (png_bytep*)malloc(sizeof(png_bytep) * height);
    for (int y = 0; y < height; y++) {
        row[y] = (png_byte*) malloc(png_get_rowbytes(png,info));
    }

    png_read_image(png, row);
    fclose(fp);

    png_destroy_read_struct(&png, &info, NULL);
}

// Сохранить картинку на диск
void imagepng::save(const char *filename) {

    int y;

    FILE *fp = fopen(filename, "wb");
    if (!fp) abort();

    png_structp png = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png) abort();

    png_infop info = png_create_info_struct(png);
    if (!info) abort();

    if (setjmp(png_jmpbuf(png))) abort();

    png_init_io(png, fp);

    // Output is 8bit depth, RGBA format.
    png_set_IHDR(
        png,
        info,
        width, height,
        8,
        PNG_COLOR_TYPE_RGBA,
        PNG_INTERLACE_NONE,
        PNG_COMPRESSION_TYPE_DEFAULT,
        PNG_FILTER_TYPE_DEFAULT
    );

    png_write_info(png, info);

    // To remove the alpha channel for PNG_COLOR_TYPE_RGB format,
    // Use png_set_filler().
    // png_set_filler(png, 0, PNG_FILLER_AFTER);

    if (!row) abort();

    png_write_image(png, row);
    png_write_end(png, NULL);

    fclose(fp);

    png_destroy_write_struct(&png, &info);
}

// Текущий пиксель
int32_t imagepng::point(int x, int y) {

    if (width == 0 || height == 0) return 0;

    x %= width;
    y %= height;

    png_bytep Y = row[y];
    png_bytep P = &(Y[x * 4]);

    return P[2] + (P[1]<<8) + (P[0]<<16) + ((255-P[3])<<24);
}

// Запись пикселя обратно
void imagepng::pset(int x, int y, int32_t cl) {

    if (width && height) {

        x %= width;
        y %= height;

        png_bytep Y = row[y];

        Y[x*4+2] = cl;           // Blue
        Y[x*4+1] = cl >> 8;      // Green
        Y[x*4+0] = cl >> 16;     // Red
        Y[x*4+3] = 255-(cl>>24); // Alpha
    }
}
